/* gethostip.c
mangels funktionierendem nslookup in 2.1.5 leistet
dieses Programm entsprechende Dienste. Auf die Dauer
wird es kleiner werden, wenn die uclibc weiter ist.
Geschrieben von Oliver Dawid, herzlichen Dank dafuer! */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <libgen.h>
#include <errno.h>
#include <resolv.h>
#include <arpa/nameser.h>
#include <malloc.h>
#include <string.h>

extern int h_errno;

# define BIGLOCK
# define BIGUNLOCK

#define MAX_RECURSE 5
#define REPLY_TIMEOUT 10
#define MAX_RETRIES 15
#define MAX_SERVERS 3
#define MAX_SEARCH 4

#define MAX_ALIASES 5

/* 1:ip + 1:full + MAX_ALIASES:aliases + 1:NULL */
#define   ALIAS_DIM   (2 + MAX_ALIASES + 1)

#ifdef DEBUG
#define DPRINTF(X,args...) fprintf(stderr, X, ##args)
#else
#define DPRINTF(X,args...)
#endif

struct resolv_answer {
		char * dotted;
		int atype;
		int aclass;
		int ttl;
		int rdlength;
		unsigned char * rdata;
		int rdoffset;
};

struct resolv_header {
		int id;
		int qr,opcode,aa,tc,rd,ra,rcode;
		int qdcount;
		int ancount;
		int nscount;
		int arcount;
};

int __nameservers;
char * __nameserver[MAX_SERVERS];
int __searchdomains;
char * __searchdomain[MAX_SEARCH];

int gethbyn_r(const char * name,
          struct hostent * result_buf,
          char * buf, size_t buflen,
          struct hostent ** result,
          int * h_errnop)
{
  struct in_addr *in;
  struct in_addr **addr_list;
  unsigned char *packet;
  struct resolv_answer a;
  int i;
  int nest = 0;
  int __nameserversXX;
  char ** __nameserverXX;
	char *buffer = buf;
	char name_buf[256];

  __open_nameservers();

  *result=NULL;
  if (!name)
    return EINVAL;

  /* do /etc/hosts first */
  if ((i=__get_hosts_byname_r(name, AF_INET, result_buf,
          buf, buflen, result, h_errnop))==0)
    return i;
  switch (*h_errnop) {
    case HOST_NOT_FOUND:
    case NO_ADDRESS:
      break;
    case NETDB_INTERNAL:
      if (errno == ENOENT) {
          break;
      }
      /* else fall through */
    default:
      return i;
  }

  DPRINTF("Nothing found in /etc/hosts\n");

  *h_errnop = NETDB_INTERNAL;
  if (buflen < sizeof(*in))
    return ERANGE;
  in=(struct in_addr*)buf;
  buf+=sizeof(*in);
  buflen-=sizeof(*in);

  if (buflen < sizeof(*addr_list)*2)
    return ERANGE;
  addr_list=(struct in_addr**)buf;
  buf+=sizeof(*addr_list)*2;
  buflen-=sizeof(*addr_list)*2;

  addr_list[0] = in;
  addr_list[1] = 0;

  if (buflen<256)
    return ERANGE;
  strncpy(buf, name, buflen);

  /* First check if this is already an address */
  if (inet_aton(name, in)) {
      result_buf->h_name = buf;
      result_buf->h_addrtype = AF_INET;
      result_buf->h_length = sizeof(*in);
      result_buf->h_addr_list = (char **) addr_list;
      *result=result_buf;
      *h_errnop = NETDB_SUCCESS;
      return NETDB_SUCCESS;
  }

  for (;;) {

  BIGLOCK;
  __nameserversXX=__nameservers;
  __nameserverXX=__nameserver;
  BIGUNLOCK;
		DPRINTF("Looking up %s\n", buf);
    i = __dns_lookup(buf, T_A, __nameserversXX, __nameserverXX, &packet, &a);

    if (i < 0) {
      *h_errnop = HOST_NOT_FOUND;
      DPRINTF("__dns_lookup\n");
      return TRY_AGAIN;
    }

    strncpy(buf, a.dotted, buflen);
    strncpy(name_buf, a.dotted, sizeof(name_buf));
    free(a.dotted);

    if (a.atype == T_CNAME) {   /* CNAME */
      DPRINTF("Got a CNAME in gethostbyname()\n");
      i = __decode_dotted(packet, a.rdoffset, buf, buflen);
      free(packet);

      if (i < 0) {
        *h_errnop = NO_RECOVERY;
        DPRINTF("__decode_dotted\n");
        return -1;
      }
      if (++nest > MAX_RECURSE) {
        *h_errnop = NO_RECOVERY;
        DPRINTF("recursion\n");
        return -1;
      }
      continue;
    } else if (a.atype == T_A) {  /* ADDRESS */
			struct resolv_header h;
      char *p = buffer;
			char **addr_ptr;
			int i, j;
			int pos = HFIXEDSZ;
			int size;

      memset(&h, 0, sizeof(h));

			__decode_header(packet, &h);

			size = (h.ancount + 1) * (sizeof(char *) + sizeof(struct in_addr)) +
						 strlen(a.dotted) + 1;

			if(buflen < size) {
				printf ("Need %d bytes but only %d are given.\n",size,buflen);
				free(a.dotted);
				free(packet);
				return ERANGE;
			}

	    for (j = 0; j < h.qdcount; j++) {
	      i = __length_question(packet, pos);
	      if (i < 0)
	        return ENOENT;
	      pos += i;
	    }

			addr_ptr = (char **)buffer;
      for (j = 0; j < h.ancount; j++) {
        i = __decode_answer(packet, pos, &a);

				if (a.atype == T_A) {
					*addr_ptr = buffer + (h.ancount + 1) * sizeof(char *) +
						          j * sizeof(struct in_addr);
					memcpy(*addr_ptr, a.rdata, sizeof(struct in_addr));
					++addr_ptr;
				}
				free(a.dotted);
				pos += i;
			}
			*addr_ptr = NULL;

			p = buffer + (h.ancount + 1) * sizeof(char *) +
				  h.ancount * sizeof(struct in_addr);
      strncpy(p, name_buf, buflen);
      result_buf->h_name = p;
			result_buf->h_aliases = NULL;
      result_buf->h_addrtype = AF_INET;
      result_buf->h_length = sizeof(struct in_addr);
			result_buf->h_addr_list = (char **) buffer;
      free(a.dotted);
      free(packet);
      break;
    } else {
      free(packet);
      *h_errnop=HOST_NOT_FOUND;
      return TRY_AGAIN;
    }
  }

  *result=result_buf;
  return NETDB_SUCCESS;
}


struct hostent *gethbyn (const char *name)
{
  static struct hostent h;
  static char buf[sizeof(struct in_addr) +
      sizeof(struct in_addr *)*2 +
      sizeof(char *)*(ALIAS_DIM) + 256/*namebuffer*/ + 32/* margin */];
  struct hostent *hp;

  DPRINTF ("Size of buffer: %d bytes\n",sizeof(buf));
  gethbyn_r(name, &h, buf, sizeof(buf), &hp, &h_errno);

  return hp;
}

int main (int argc, char **argv)
{
  char* p = basename (argv[0]);

  struct hostent* h = NULL;
  int i;
  int j;
  struct in_addr inp;

  if (!strcmp(p, "gethostip")) {
    if (argc != 2) {
      puts ("\nUsage: gethostip <name>\n");
      return 1;
    }
    h = gethostbyname(argv[1]);
    h = gethbyn(argv[1]);

    if (!h)
      return 1;
    for (j = 0; h->h_addr_list[j]; j++) {
      for (i = 0; i < h->h_length; ++i) {
        printf ("%d", (unsigned char)h->h_addr_list[j][i]);
        if (i < h->h_length - 1)
          putchar ('.');
      }
      puts ("");
    }
    return 0;
  } else if (!strcmp(p, "gethostname")) {
    if (argc != 2) {
      puts ("\nUsage: gethostname <name>\n");
      return 1;
    }
    if (!inet_aton (argv[1], &inp))
      return 1;
    h = gethostbyaddr((char *)&inp.s_addr, sizeof(inp.s_addr), AF_INET);

    if (!h)
      return 1;
    if (h->h_name)
      puts (h->h_name);
    return 0;
  }
}




